/*
 * WARNING! This file is automatically generated - DO NOT EDIT!
 */
#define STATE 0
#define COUNTER 4
#define BLOCKED 16
#define SIGNAL 12
#define KERNEL_STACK_PAGE 88
#define TSS 504
#define KSP 0
#define MMU_SEG0 8
#define MMU_SEG1 12
#define MMU_SEG2 16
#define MMU_SEG3 20
#define MMU_SEG4 24
#define MMU_SEG5 28
#define MMU_SEG6 32
#define MMU_SEG7 36
#define MMU_SEG8 40
#define MMU_SEG9 44
#define MMU_SEG10 48
#define MMU_SEG11 52
#define MMU_SEG12 56
#define MMU_SEG13 60
#define MMU_SEG14 64
#define MMU_SEG15 68
#define TSS_FPR0 72
#define TSS_FPR1 80
#define TSS_FPR2 88
#define TSS_FPR3 96
#define TSS_FPR4 104
#define TSS_FPR5 112
#define TSS_FPR6 120
#define TSS_FPR7 128
#define TSS_FPR8 136
#define TSS_FPR9 144
#define TSS_FPR10 152
#define TSS_FPR11 160
#define TSS_FPR12 168
#define TSS_FPR13 176
#define TSS_FPR14 184
#define TSS_FPR15 192
#define TSS_FPR16 200
#define TSS_FPR17 208
#define TSS_FPR18 216
#define TSS_FPR19 224
#define TSS_FPR20 232
#define TSS_FPR21 240
#define TSS_FPR22 248
#define TSS_FPR23 256
#define TSS_FPR24 264
#define TSS_FPR25 272
#define TSS_FPR26 280
#define TSS_FPR27 288
#define TSS_FPR28 296
#define TSS_FPR29 304
#define TSS_FPR30 312
#define TSS_FPR31 320
#define INT_FRAME_SIZE 384
#define GPR0 56
#define GPR1 60
#define GPR2 64
#define GPR3 68
#define GPR4 72
#define GPR5 76
#define GPR6 80
#define GPR7 84
#define GPR8 88
#define GPR9 92
#define GPR10 96
#define GPR11 100
#define GPR12 104
#define GPR13 108
#define GPR14 112
#define GPR15 116
#define GPR16 120
#define GPR17 124
#define GPR18 128
#define GPR19 132
#define GPR20 136
#define GPR21 140
#define GPR22 144
#define GPR23 148
#define GPR24 152
#define GPR25 156
#define GPR26 160
#define GPR27 164
#define GPR28 168
#define GPR29 172
#define GPR30 176
#define GPR31 180
#define FPR0 256
#define FPR1 264
#define FPR2 272
#define FPR3 280
#define FPCSR 288
#define _NIP 184
#define _MSR 188
#define _CTR 192
#define _LINK 196
#define _CCR 200
#define _XER 208
#define _DAR 212
#define _DSISR 216
#define _HASH1 220
#define _HASH2 224
#define _IMISS 228
#define _DMISS 232
#define _ICMP 236
#define _DCMP 240
#define ORIG_GPR3 244
#define RESULT 248
#define TRAP 296
#define MARKER 300
