/* $Id: string.h,v 1.15 1995/11/25 02:32:56 davem Exp $
 * string.h: External definitions for optimized assembly string
 *           routines for the Linux Kernel.
 *
 * Copyright (C) 1995 David S. Miller (davem@caip.rutgers.edu)
 */

#ifndef __SPARC_STRING_H__
#define __SPARC_STRING_H__

/* Nothing yet... */

#endif /* !(__SPARC_STRING_H__) */
