/*
 *		IP_ALIAS (AF_INET) aliasing definitions.
 *
 *
 * Version:	@(#)ip_alias.h	0.43   12/20/95
 *
 * Author:	Juan Jose Ciarlante, <jjciarla@raiz.uncu.edu.ar>
 *
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *	
 */

#ifndef _IP_ALIAS_H
#define _IP_ALIAS_H

extern int ip_alias_init(void);
extern int ip_alias_done(void);

#endif  /* _IP_ALIAS_H */
